""" file:   __init__.py (pysiss.vocabulary.erml)
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: init file
"""

import unmarshallers

__all__ = [unmarshallers]
