""" file: __init__.py (pysiss.borehole.properties)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Imports for pysiss.borehole.properties
"""

from .property import Property
from .property_type import PropertyType